# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
from .api_handler import *
from .data_sources import *
from .props import *
from .prop_types import *
from .logical_operators import *



# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
from doughboy.api_handler import *
import doughboy.props as props
import doughboy.prop_types as prop_types

class data_source_page:
    def __init__(self):
        self.payload:dict = None
        self.url = prop_types.url_prop("url")
        self.icon = prop_types.icon_prop("icon")
        self.data_source_id:str = None
        self.template:str = None

    def set_payload(self, payload:dict=None):
        if payload is None:
            return

        self.payload = payload
        self.id = payload["id"]
        self.icon.payload = { "icon": payload["icon"] }
        self.url.payload = { "url": payload["url"] }
        self.data_source_id = payload["parent"]["data_source_id"]

class data_source:
    __data_source_id__:str = None
    __template__:str = None

    @classmethod
    def create_page_object(cls) -> data_source_page:
        new_page_object:data_source_page = data_source_page()

        for property_name, property_value in cls.__dict__.items():
            if isinstance(property_value, props.props):
                setattr(new_page_object, property_name, property_value.property_type_class(property_name, {}))
        new_page_object.data_source_id = cls.__data_source_id__
        new_page_object.template = cls.__template__

        return new_page_object
    
class describer:
    def __init__(self):
        self.title:str = None

    def __str__(self) -> str:
        messages = [
            f"{self.title}:",
            "",
            f"{'property name':32} {'type':16} {'id':8}",
            "-" * 32 + " " + "-" * 16 + " " + "-" * 8,
        ]
        for attribute_name in dir(self):
            if attribute_name.startswith("__"):
                continue

            if attribute_name == "title":
                continue

            value:dict = getattr(self, attribute_name)
            messages.append(f"{attribute_name:32} {value['type']:16} {value['id']:8}")

        return "\n".join(messages)

class doughboy:
    def __init__(self, api_key:str):
        self.api_handler:notion_api_handler = notion_api_handler(api_key)
        self.file_uploader:file_uploader = file_uploader(self.api_handler)
        self.selection_columns:list = None
        self.filter:dict = None

    def select_from(self, *args) -> object:
        return selector(self, *args)

    def __is_data_source_class(self, data_source_class:data_source) -> bool:
        return False if not data_source_class or not issubclass(data_source_class, data_source) else True

    def update_to(self, data_source_class:data_source) -> object:
        if not self.__is_data_source_class(data_source_class):
            raise ValueError("exactly one data_source_class is required")

        return updator(self, data_source_class)

    def delete_from(self, data_source_class:data_source) -> object:
        if not self.__is_data_source_class(data_source_class):
            raise ValueError("exactly one data_source_class is required")

        return deletor(self, data_source_class)

    def insert_into(self, data_source_class:data_source) -> object:
        if not self.__is_data_source_class(data_source_class):
            raise ValueError("exactly one data_source_class is required")

        return insertor(self, data_source_class)

    def __is_page_object(self, page_object) -> bool:
        return False if not page_object or not isinstance(page_object, data_source_page) else True

    def update_one(self, page_object:data_source_page=None) -> object:
        if not self.__is_page_object(page_object):
            raise ValueError("page_object is required")

        payload:dict = {
            "properties": {}
        }
        target_properties:dict = { key: value for key, value in page_object.__dict__.items() if isinstance(value, prop_types.prop_type_base) and value.value_updated }
        if not target_properties:   # all properties is up to date. skip update.
            if page_object.property_updated == False:
                return None
        else:
            for value in target_properties.values():
                if value.type_name not in [ "icon", "url" ]:
                    payload["properties"].update(value.to_dict())
                else:
                    if value.value is not None:
                        payload.update({ value.type_name: value.value })

        response = self.api_handler.patch(f"pages/{page_object.id}", payload)
        [ setattr(value, "value_updated", False) for value in target_properties.values() ]
        return response

    def delete_one(self, page_object:data_source_page=None) -> object:
        if not self.__is_page_object(page_object):
            raise ValueError("page_object is required")

        payload:dict = {
            "id": page_object.id,
            "archived": True,
        }

        return self.api_handler.patch(f"pages/{page_object.id}", payload)

    def insert_one(self, page_object:data_source_page=None) -> object:
        if not self.__is_page_object(page_object):
            raise ValueError("page_object is required")

        # clear page property if selected page    
        if page_object.payload is not None and "id" in page_object.payload:
            page_object.url.value = None
            page_object.icon.value = None

        payload:dict = {
            "parent": {
                "data_source_id": page_object.data_source_id,
            },
            "properties": {}
        }
        if page_object.template is not None:
            if page_object.template == "default":
                payload["template"] = { "type": "default" }
            else:
                payload["template"] = { "type": "template_id", "template_id": page_object.template }

        target_properties:dict = { key: value for key, value in page_object.__dict__.items() if isinstance(value, prop_types.prop_type_base) }
        if not target_properties:
            raise ValueError("page_object has no properties")

        for value in target_properties.values():
            if value.type_name not in [ "icon", "url" ]:
                if value.value is not None:
                    payload["properties"].update(value.to_dict())           # data source property
            else:
                if value.value is not None:
                    payload.update({ value.type_name: value.value })        # page property

        response = self.api_handler.post("pages", payload)
        page_object.id = response.get("id")
        [ setattr(value, "value_updated", False) for value in target_properties.values() ]

        page_object.set_payload(response)

        return page_object

    def describe(self, data_source_class:data_source):
        if not self.__is_data_source_class(data_source_class):
            raise ValueError("exactly one data_source_class is required")

        describe_object:describer = describer()
        response = self.api_handler.get(f"data_sources/{data_source_class.__data_source_id__}")
        describe_object.title = "".join([ item["text"]["content"] for item in response["title"] ]) if len(response["title"]) > 0 else None
        for key, value in response["properties"].items():
            setattr(describe_object, key, { "type": value["type"], "id": value["id"] })

        return describe_object

class accessor:
    def __init__(self, controller:doughboy):
        self.controller:doughboy = controller
        self.data_source_class:data_source = None

class selector(accessor):
    def __init__(self, controller:doughboy, *args):
        super().__init__(controller)
        self.data_source_class:data_source = None
        self.selection_properties:list = None
        self.filter:dict = None
        self.sorts:dict = None
        self.pickup_selection_properties(*args)

    def pickup_selection_properties(self, *args) -> None:
        if len(args) < 1:
            raise ValueError("at least one data_source_class is required")

        if isinstance(args[0], props.prop_descriptor):
            data_sources:set = (arg.property.data_source_class for arg in args)
            if len(data_sources) > 1:
                raise ValueError("all properties must belong to the same data_source_class")

            self.data_source_class = args[0].data_source_class
            self.selection_properties = [arg.property.__name__ for arg in args]
        elif issubclass(args[0], data_source):
            self.data_source_class = args[0]
            self.selection_properties = [prop.__name__ for prop in args[0].__dict__.values() if isinstance(prop, props.props)]
        else:
            raise ValueError("invalid properties selection arguments")

    def where(self, *args) -> accessor:
        if len(args) < 1:
            return self

        self.filter = args[0]
        return self

    def order_by(self, *args) -> accessor:
        if len(args) < 1:
            return self

        self.sorts = [ *args ]
        return self

    def exec(self, filter_title_properties=False) -> list:
        payload:dict = {}
        if self.filter:
            payload["filter"] = self.filter

        if self.sorts:
            payload["sorts"] = self.sorts

        pages = []
        while True:
            url = f"data_sources/{self.data_source_class.__data_source_id__}/query"
            if filter_title_properties:
                url += "?filter_properties[]=title"

            response = self.controller.api_handler.post(url, payload)
            pages.extend(response.get("results", []))
            if not response.get("has_more"):
                break

        new_page_objects:list = []
        for page in pages:
            new_page_object:data_source_page = data_source_page()
            new_page_object.set_payload(page)

            for property_name, property_value in page["properties"].items():
                if property_name not in self.selection_properties:
                    continue

                property:props.props = self.data_source_class.__dict__.get(property_name)
                property_type_class:prop_types.prop_type_base = property.property_type_class(property_name, property_value)
                property_type_class.parent = self.controller
                setattr(new_page_object, property_name, property_type_class)
            new_page_objects.append(new_page_object)

        return new_page_objects

class updator(accessor):
    def __init__(self, controller:doughboy, data_source_class:data_source):
        super().__init__(controller)
        self.data_source_class:data_source = data_source_class
        self.selector:selector = selector(controller, (data_source_class))

    def where(self, *args) -> accessor:
        if len(args) < 1:
            return self

        self.selector = self.selector.where(*args)
        return self

    def values(self, **kwargs) -> None:
        if len(kwargs) < 1:
            raise ValueError("at least one property is required")

        for page in self.selector.exec():
            self.update_page_object(page, **kwargs)
            self.controller.update_one(page)

    def update_page_object(self, page_object, **kwargs):
        for key, value in kwargs.items():
            if hasattr(page_object, key):
                property:object = getattr(page_object, key)
                if isinstance(property, prop_types.prop_type_base):
                    property.value = value
                else:
                    raise ValueError(f"{key} is not a property")
            else:
                raise ValueError(f"{key} is not a property of {self.data_source_class.__name__}")

class deletor(accessor):
    def __init__(self, controller:doughboy, data_source_class:data_source):
        super().__init__(controller)
        self.selector = selector(controller, (data_source_class))

    def where(self, *args) -> accessor:
        if len(args) < 1:
            return self

        self.selector = self.selector.where(*args)
        return self

    def exec(self) -> None:
        pages:list = self.selector.exec(filter_title_properties=True)
        [ self.controller.delete_one(page) for page in pages ]

class insertor(accessor):
    def __init__(self, controller:doughboy, data_source_class:data_source):
        super().__init__(controller)
        self.data_source_class:data_source = data_source_class

    def values(self, **kwargs) -> data_source_page:
        if len(kwargs) < 1:
            raise ValueError("at least one property is required")

        new_page_object:data_source_page = self.data_source_class.create_page_object()
        for key, value in kwargs.items():
            if hasattr(new_page_object, key):
                property:prop_types.prop_type_base = getattr(new_page_object, key)
                if isinstance(property, prop_types.prop_type_base):
                    property.value = value
                    property.parent = self.controller
                else:
                    raise ValueError(f"{key} is not a property")
            else:
                raise ValueError(f"{key} is not a property of {self.data_source_class.__name__}")

        self.controller.insert_one(new_page_object)
        return new_page_object



# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
def or_(*filters) -> dict:
    return { "or": [ filter for filter in filters ] }

def and_(*filters) -> dict:
    return { "and": [ filter for filter in filters ] }

def asc_(descriptor) -> dict:
    return { "property": descriptor.property.__name__, "direction": "ascending" }

def ascending(descriptor) -> dict:
    return asc_(descriptor)

def desc_(descriptor) -> dict:
    return { "property": descriptor.property.__name__, "direction": "descending" }

def descending(descriptor) -> dict:
    return desc_(descriptor)



# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
import urllib
import os

class prop_type_base:
    def __init__(self, name:str, payload:dict={}):
        self.name:str = name
        self.payload:dict = payload
        self.value_updated:bool = False
        self.parent = None

    @property
    def value(self) -> dict:
        raise NotImplementedError("Subclasses must implement value property")

    @value.setter
    def value(self, val) -> None:
        raise NotImplementedError("Subclasses must implement value setter")

    @property
    def default_value(self) -> dict:
        raise NotImplementedError("Subclasses must implement empty_value method")

    def to_dict(self) -> dict:
        return { self.name: self.payload }

class title_prop(prop_type_base):
    type_name:str = "title"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "title": [ { "type": "text", "text": { "content": "" }, "plain_text": "" } ] }
    
    @property
    def value(self) -> str:
        if "title" not in self.payload:
            self.payload.update(self.default_value)

        if len(self.payload["title"]) == 0:
            return None

        return "".join([ item["text"]["content"] for item in self.payload["title"] ])

    @value.setter
    def value(self, value) -> None:
        if value == self.value:
            return

        if value is None:
            self.payload["title"] = []
        else:
            if len(self.payload["title"]) == 0:
                self.payload.update(self.default_value)
            else:
                self.payload["title"] = self.payload["title"][:1]

            self.payload["title"][0]["text"]["content"] = value
            self.payload["title"][0]["plain_text"] = value
        self.value_updated = True

class rich_text_prop(prop_type_base):
    type_name:str = "rich_text"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)
   
    @property
    def default_value(self) -> dict:
        return { "rich_text": [ { "type": "text", "text": { "content": "" }, "plain_text": "" } ] }

    @property
    def value(self) -> str:
        if "rich_text" not in self.payload:
            self.payload.update(self.default_value)

        if len(self.payload["rich_text"]) == 0:
            return None

        return "".join([ item["text"]["content"] for item in self.payload["rich_text"] ])

    @value.setter
    def value(self, value) -> None:
        if value == self.value:
            return

        if value is None:
            self.payload["rich_text"] = []
        else:
            if len(self.payload["rich_text"]) == 0:
                self.payload.update(self.default_value)
            else:
                self.payload["rich_text"] = self.payload["rich_text"][:1]

            self.payload["rich_text"][0]["text"]["content"] = value
            self.payload["rich_text"][0]["plain_text"] = value
        self.value_updated = True

class date_prop(prop_type_base):
    type_name:str = "date"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "date": { "start": None, "end": None, "time_zone": None } }

    @property
    def value(self) -> dict:
        if "date" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["date"]

    @value.setter
    def value(self, values:dict) -> None:
        if values is None:
            if self.value is None:
                return

            self.payload["date"] = None
            self.value_updated = True
            return
        else:
            if self.value is not None:
                if not set(values) ^ set(self.value):
                    return

            self.payload.update(self.default_value)

        if "start" in values:
            self.start = values["start"]

        if "end" in values:
            self.end = values["end"]

        if "time_zone" in values:
            self.time_zone = values["time_zone"]

    def __normalize_payload(self):
        if self.start is None:
            if self.end is None and self.time_zone is None:
                self.payload["date"] = None
                return
            
            raise ValueError("must set property 'start', if configure property 'end' or 'time_zone'.")

    @property
    def start(self) -> str:
        if "date" not in self.payload or self.payload["date"] is None:
            self.payload.update(self.default_value)

        return self.payload["date"]["start"]

    @start.setter
    def start(self, value:str) -> None:
        if value == self.start:
            return

        self.payload["date"]["start"] = value
        self.__normalize_payload()
        self.value_updated = True

    @property
    def end(self) -> str:
        if "date" not in self.payload or self.payload["date"] is None:
            self.payload.update(self.default_value)

        return self.payload["date"]["end"]

    @end.setter
    def end(self, value:str) -> None:
        if value == self.end:
            return

        self.payload["date"]["end"] = value
        self.__normalize_payload()
        self.value_updated = True

    @property
    def time_zone(self) -> str:
        if "date" not in self.payload or self.payload["date"] is None:
            self.payload.update(self.default_value)

        return self.payload["date"]["time_zone"]
    
    @time_zone.setter
    def time_zone(self, value:str) -> None:
        if value == self.value:
            return

        self.payload["date"]["time_zone"] = value
        self.__normalize_payload()
        self.value_updated = True

class number_prop(prop_type_base):
    type_name:str = "number"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "number": None }

    @property
    def value(self) -> float:
        if "number" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["number"]

    @value.setter
    def value(self, value:float) -> None:
        if value == self.value:
            return

        self.payload["number"] = value
        self.value_updated = True

class select_prop(prop_type_base):
    type_name:str = "select"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "select": None }

    @property
    def value(self) -> str:
        if "select" not in self.payload:
            self.payload.update(self.default_value)

        if self.payload["select"] is None:
            return None

        return self.payload["select"]["name"]

    @value.setter
    def value(self, value:str) -> None:
        if value == "":
            raise ValueError("null-string is can not set to property type 'select'.")

        if value == self.value:
            return

        if value is None:
            self.payload["select"] = None
            self.value_updated = True
            return

        if self.payload["select"] is None:
            self.payload["select"] = { "name": value }
        else:
            self.payload["select"].update({ "name": value })
        self.value_updated = True

class multi_select_prop(prop_type_base):
    type_name:str = "multi_select"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "multi_select": [] }

    @property
    def value(self) -> list:
        if "multi_select" not in self.payload:
            self.payload.update(self.default_value)

        if not self.payload["multi_select"]:
            return self.payload["multi_select"]

        return [ item["name"] for item in self.payload["multi_select"] ]

    @value.setter
    def value(self, values:list) -> None:
        if values is None:
            raise ValueError("None is can not set to property type 'multi_select'.")

        if not set(values) ^ set(self.value):
            return

        self.payload["multi_select"] = [ { "name": value } for value in values ]
        self.value_updated = True

    def replace(self, old_name, new_name):
        try:
            index = self.value.index(old_name)
            self.payload["multi_select"][index] = { "name": new_name }
            self.value_updated = True
        except:
            raise KeyError(f"multi-select name {old_name} is not in list.")

    def remove(self, name):
        try:
            index = self.value.index(name)
            del self.payload["multi_select"][index]
        except Exception as e:
            raise KeyError(f"multi-select name {id} is not in list.")

    def append(self, name):
        self.payload["multi_select"].append({ "name": name })

class status_prop(prop_type_base):
    type_name:str = "status"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "status": None }

    @property
    def value(self) -> str:
        if "status" not in self.payload:
            self.payload.update(self.default_value)

        if self.payload["status"] is None:
            return None

        return self.payload["status"]["name"]

    @value.setter
    def value(self, value:str) -> None:
        if value == "":
            raise ValueError("null-string is can not set to property type 'status'.")

        if value == self.value:
            return

        if value is None:
            self.payload["status"] = None
            self.value_updated = True
            return

        self.payload["status"] = { "name": value }
        self.value_updated = True

class people_prop(prop_type_base):
    type_name:str = "people"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "people": [] }

    @property
    def value(self) -> list:
        if "people" not in self.payload:
            self.payload.update(self.default_value)

        return [ item["id"] for item in self.payload["people"] ]

    @value.setter
    def value(self, values:list) -> None:
        if values is None:
            raise ValueError("None is can not set to property type 'people'.")

        if not set(values) ^ set(self.value):
            return

        self.payload["people"] = [ { "id": value } for value in values ]
        self.value_updated = True

    def replace(self, old_id, new_id):
        try:
            index = self.value.index(old_id)
            self.payload["people"][index] = { "id": new_id }
            self.value_updated = True
        except Exception as e:
            raise KeyError(f"people id {old_id} is not in list.")

    def remove(self, id):
        try:
            index = self.value.index(id)
            del self.payload["people"][index]
        except Exception as e:
            raise KeyError(f"people id {id} is not in list.")

    def append(self, id):
        self.payload["people"].append({ "id": id })

class checkbox_prop(prop_type_base):
    type_name:str = "checkbox"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "checkbox": False }

    @property
    def value(self) -> bool:
        if "checkbox" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["checkbox"]

    @value.setter
    def value(self, value:bool) -> None:
        if value is None:
            raise ValueError("None is can not set to property type 'checkbox'.")

        if value == self.value:
            return

        self.payload["checkbox"] = value
        self.value_updated = True

class url_prop(prop_type_base):
    type_name:str = "url"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "url": None }

    @property
    def value(self) -> str:
        if "url" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["url"]

    @value.setter
    def value(self, value:str) -> None:
        if value == "":
            raise ValueError("null-string is can not set to property type 'url'.")

        if value == self.value:
            return

        self.payload["url"] = value
        self.value_updated = True

class email_prop(prop_type_base):
    type_name:str = "email"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "email": None }

    @property
    def value(self) -> str:
        if "email" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["email"]

    @value.setter
    def value(self, value:str) -> None:
        if value == "":
            raise ValueError("null-string is can not set to property type 'email'.")

        if value == self.value:
            return

        self.payload["email"] = value
        self.value_updated = True

class phone_number_prop(prop_type_base):
    type_name:str = "phone_number"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "phone_number": None }

    @property
    def value(self) -> str:
        if "phone_number" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["phone_number"]

    @value.setter
    def value(self, value:str) -> None:
        if value == "":
            raise ValueError("null-string is can not set to property type 'phone_number'.")

        if value == self.value:
            return

        self.payload["phone_number"] = value
        self.value_updated = True

class relation_prop(prop_type_base):
    type_name:str = "relation"

    def __init__(self, name:str, payload:dict={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "relation": [] }

    @property
    def value(self) -> list:
        if "relation" not in self.payload:
            self.payload.update(self.default_value)

        return [ item["id"] for item in self.payload["relation"] ]

    @value.setter
    def value(self, values:list) -> None:
        if values is None:
            raise ValueError("None is can not set to property type 'relation'.")

        if not set(values) ^ set(self.value):
            return

        self.payload["relation"] = [ { "id": value } for value in values ]
        self.value_updated = True

class file_prop(prop_type_base):
    type_name:str = "files"

    def __init__(self, name, payload={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "files": [] }

    def __make_file_check_list(self, values:list) -> list:
        items:list = []
        for value in values:
            for key, val in value.items():
                if key in [ "file", "external" ]:
                    items.append(f"{key}_{urllib.parse.quote(val['url'])}")
                elif key == "file_upload":
                    items.append(f"{key}_{val['id']}")

        return items

    @property
    def value(self) -> list:
        if "files" not in self.payload:
            self.payload.update(self.default_value)

        if len(self.payload["files"]) == 0:
            return []

        files:list = []
        for file in self.payload["files"]:
            if "file" in file:
                files.append({ "file": file["file"] })
            elif "external" in file:
                files.append({ "external": file["external"] })
            else:
                files.append({ "file_upload": file["file_upload"] })

        return files

    @value.setter
    def value(self, values:list) -> None:
        if values is None:
            raise ValueError("None is can not set to property type 'file'.")

        if not set(self.__make_file_check_list(values)) ^ set(self.__make_file_check_list(self.value)):
            return

        quoted_files:list = []
        for value in values:
            if "file" in value:
                if "url" not in value["file"]:
                    raise ValueError("ui uploaded file is must set url property.")
                
                value["file"]["url"] = urllib.parse.quote(value["file"]["url"])
            elif "external" in value:
                if "url" not in value["external"]:
                    raise ValueError("external file is must set url property.")

                value["external"]["url"] = urllib.parse.quote(value["external"]["url"])
            elif "file_upload" in value:
                if "id" not in value["file_upload"]:
                    raise ValueError("api uploaded file is must set id property.")
            else:
                raise ValueError(f"unknown file type: {value.keys()}.")

            quoted_files.append(value)

        self.payload["files"] = quoted_files
        self.value_updated = True

    def append(self, file_info:dict) -> None:
        self.payload["files"].append(file_info)
        self.value_updated = True

    def upload(self, filename:str) -> None:
        if os.path.getsize(filename) > self.parent.file_uploader.chunk_size:
            self.append(self.parent.file_uploader.multipart_upload(filename))
        else:
            self.append(self.parent.file_uploader.singlepart_upload(filename))

    def import_external_file(self, external_url:str, filename:str) -> None:
        self.append(self.parent.file_uploader.import_external_file(external_url, filename))

class icon_prop(prop_type_base):
    type_name:str = "icon"

    def __init__(self, name, payload={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "icon": None }

    @property
    def value(self) -> str:
        if "icon" not in self.payload:
            self.payload.update(self.default_value)

        if self.payload["icon"] is None:
            return None

        if "file" in self.payload["icon"]:
            return self.file
        elif "external" in self.payload["icon"]:
            return self.external

        return self.emoji

    @value.setter
    def value(self, value:dict) -> None:
        if value is None:
            if self.value is not None:
                self.value_updated = True
            self.payload["icon"] = None
            return

        if set(value) ^ set(self.value):
            return

        if "file" in value:
            self.file = value["file"]
        elif "eternal" in value:
            self.external = value["external"]
        elif "emoji" in value:
            self.emoji = value["emoji"]
        else:
            raise ValueError(f"unknown icon type: {value.keys()}.")

    @property
    def file(self) -> str:
        if "icon" not in self.payload:
            self.payload.update(self.default_value)

        if "file" not in self.payload["icon"]:
            return None

        return self.payload["icon"]["file"]["url"]

    @file.setter
    def file(self, value:str) -> None:
        if set(value) ^ set(self.file):
            return

        if value is None:
            self.payload["icon"] = None
            return
        
        self.payload["icon"].update({ "file": { "url": value } })
        self.value_updated = True

    @property
    def external(self) -> str:
        if "icon" not in self.payload:
            self.payload.update(self.default_value)

        if "external" not in self.payload["icon"]:
            return None

        return self.payload["icon"]["external"]["url"]

    @external.setter
    def external(self, value:str) -> None:
        if set(value) ^ set(self.external):
            return

        if value is None:
            self.payload["icon"] = None
            return
        
        self.payload["icon"].update({ "external": { "url": value } })
        self.value_updated = True

    @property
    def emoji(self) -> str:
        if "icon" not in self.payload:
            self.payload.update(self.default_value)

        if "emoji" not in self.payload["icon"]:
            return None

        return self.payload["icon"]["emoji"]

    @emoji.setter
    def emoji(self, value:str) -> None:
        if set(value) ^ set(self.emoji):
            return

        if value is None:
            self.payload["icon"] = None
            return
        
        self.payload["icon"].update({ "emoji": value })
        self.value_updated = True

class url_prop(prop_type_base):
    type_name:str = "url"

    def __init__(self, name, payload={}):
        super().__init__(name, payload)

    @property
    def default_value(self) -> dict:
        return { "url": None }

    @property
    def value(self) -> str:
        if "url" not in self.payload:
            self.payload.update(self.default_value)

        return self.payload["url"]

    @value.setter
    def value(self, value:dict) -> None:
        if value == self.value:
            return

        self.payload["url"] = value
        self.value_updated = True



# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
import inspect

class props:
    def __init__(self, property_name:str, property_type_class:object):
        self.__name__:str = property_name
        self.property_type_class:object = property_type_class

    def __get__(self, _, owner:object):
        return prop_descriptor(self, owner)

operator_property_type_map = {
    # ex.)
    # operator_name : [ property-type, ... ]

    # python operators
    "__ne__": [
        "checkbox",
        "date",
        "id",
        "number",
        "rich_text",
        "status",
        "title",
    ],
    "__eq__": [
        "checkbox",
        "date",
        "id",
        "number",
        "rich_text",
        "status",
        "title",
    ],
    "__gt__": [
        "date",
        "id",
        "number",
    ],
    "__ge__": [
        "date",
        "id",
        "number",
    ],
    "__lt__": [
        "date",
        "id",
        "number",
    ],
    "__le__": [
        "date",
        "id",
        "number",
    ],

    # notion api operators
    "after": [
    	"date",
    ],
    "any": [
        "rollup",
    ],
    "before": [
        "date",
    ],
    "contains": [
        "multi-select",
        "people",
        "relation",
        "rich_text",
        "select",
        "title",
    ],
    "does_not_contain": [
        "multi-select",
        "people",
        "relation",
        "rich_text",
        "select",
        "title",
    ],
    "does_not_equals": [
        "checkbox",
        "id",
        "number",
        "rich_text",
        "status",
        "title",
    ],
    "ends_with": [
        "rich_text",
        "title",
    ],
    "equals": [
        "checkbox",
        "date",
        "id",
        "number",
        "rich_text",
        "status",
        "title",
    ],
    "every": [
        "rollup",
    ],
    "greater_than": [
        "id",
        "number",
    ],
    "greater_than_or_equal_to": [
        "id",
        "number",
    ],
    "is_empty": [
        "date",
        "files",
        "multi-select",
        "number",
        "poeple",
        "relaton",
        "rich_text",
        "select",
        "status",
        "title",
   ],
    "is_not_empty": [
        "date",
        "files",
        "multi-select",
        "number",
        "poeple",
        "relaton",
        "rich_text",
        "select",
        "status",
        "title",
    ],
    "less_than": [
        "id",
        "number",
    ],
    "less_than_or_equal_to": [
        "id",
        "number",
    ],
    "next_month": [
        "date",
    ],
    "next_week": [
        "date",
    ],
    "next_year": [
        "date",
    ],
    "none": [
        "rollup",
    ],
    "on_or_after": [
        "date",
    ],
    "on_or_before": [
        "date",
    ],
    "past_month": [
        "date",
    ],
    "past_week": [
        "date",
    ],
    "past_year": [
        "date",
    ],
    "starts_with": [
        "rich_text",
        "title",
    ],
    "this_week": [
        "date",
    ],
}

operator_conversion_map = {
    "__ne__": "does_not_equal",
    "__eq__": "equals",
    "__gt__": "greater_than",
    "__ge__": "greater_than_or_equal_to",
    "__lt__": "less_than",
    "__le__": "less_than_or_equal_to",
}

class  prop_descriptor:
    def __init__(self, property:props, data_source_class:object):
        self.property:props = property
        self.data_source_class:object = data_source_class

    def set(self, value:str):
        return { self.property.__name__: value }

    # Generic operator call
    def __operator_call(self, operator:str, other:object):
        if self.property.property_type_class.type_name not in operator_property_type_map[operator]:
            raise ValueError(f"{operator} is not available for {self.property.property_type_class.type_name}")

        #replace python operators to notion api operators
        if operator in operator_conversion_map:
            if self.property.property_type_class.type_name == "date":
                if operator == "__gt__":
                    operator = "after"
                elif operator == "__ge__":
                    operator = "on_or_after"
                elif operator == "__lt__":
                    operator = "before"
                elif operator == "__le__":
                    operator = "on_or_before"
            else:
                operator = operator_conversion_map[operator]

        return {
            "property": self.property.__name__,
            self.property.property_type_class.type_name: {
                operator: other
            }
        }

    # python operators
    def __eq__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def __ge__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def __gt__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def __le__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def __lt__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def __ne__(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    # notion api operators
    def after(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def any(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def before(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def contains(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def does_not_contain(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def does_not_equal(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def ends_with(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def equals(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def every(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def greater_than(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def greater_than_or_equal_to(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def is_empty(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, True)

    def is_not_empty(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, True)

    def less_than(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def less_than_or_equal_to(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def next_month(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})
    
    def next_week(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})
    
    def next_year(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})

    def none(self, other):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def on_or_after(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)

    def on_or_before(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def past_month(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})
    
    def past_week(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})
    
    def past_year(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})

    def starts_with(self, other:object):
        return self.__operator_call(inspect.currentframe().f_code.co_name, other)
    
    def this_week(self):
        return self.__operator_call(inspect.currentframe().f_code.co_name, {})



# doughboy
# https://github.com/Hajime-Saitou/doughboy
#
# Copyright (c) 2025 Hajime Saito
# MIT License
import requests
import json
import math
import time
import os

class notion_api_handler:
    def __init__(self, api_key:str, api_version:str = "2025-09-03", base_url:str = "https://api.notion.com/v1/"):
        if not api_key:
            raise ValueError("API key is required")

        if api_key.startswith("secret_"):
            raise ValueError("Please use the new API key format.")

        if not api_key.startswith("ntn_"):
            raise ValueError("Invalid API key format")

        self.api_key:str = api_key
        self.api_version:str = api_version
        self.base_url:str = base_url

    def make_headers(self, content_type=None) -> dict:
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Notion-Version": self.api_version
        }
        if content_type:
            headers["Content-Type"] = content_type

        return headers

    def make_endpoint_url(self, endpoint:str) -> str:
        return f"{self.base_url}{endpoint}"

    def get(self, endpoint:str, params=None, headers:dict=None) -> dict:
        if not headers:
            headers = self.make_headers("application/json")

        response = requests.get(self.make_endpoint_url(endpoint), headers=headers, params=params)
        response.raise_for_status()
        return response.json()

    def post(self, endpoint:str, data:dict={}, headers:dict=None) -> dict:
        if not headers:
            headers = self.make_headers("application/json")

        response = requests.post(self.make_endpoint_url(endpoint), headers=headers, data=json.dumps(data))
        response.raise_for_status()
        return response.json()

    def patch(self, endpoint:str, data:dict={}, headers:dict=None) -> dict:
        if not headers:
            headers = self.make_headers("application/json")

        response = requests.patch(self.make_endpoint_url(endpoint), headers=headers, data=json.dumps(data))
        response.raise_for_status()
        return response.json()

    def delete(self, endpoint:str, headers:dict=None) -> bool:
        if not headers:
            headers = self.make_headers("application/json")

        response = requests.delete(self.make_endpoint_url(endpoint), headers=headers)
        response.raise_for_status()
        return response.status_code == 204

# https://developers.notion.com/docs/working-with-files-and-media
class file_uploader:
    def __init__(self, api_handler):
        self.api_handler:notion_api_handler = api_handler
        self.chunk_size:int = 20 * 1024 * 1024

    def get_upload_id(self, data={}):
        headers:dict = self.api_handler.make_headers()
        response = self.api_handler.post("file_uploads", data=data, headers=headers)
        return response["id"]
    
    def singlepart_upload(self, filename:str) -> dict:
        basefilename:str = os.path.basename(filename)
        upload_id:str = self.get_upload_id()

        headers:dict = self.api_handler.make_headers()
        endpoint_url:str = self.api_handler.make_endpoint_url(f"file_uploads/{upload_id}/send")
        with open(filename, "rb") as f:
            files = {
                "file": (basefilename, f),
                "part_number": (None, "1")
            }

            response = requests.post(endpoint_url, headers=headers, files=files)
            response.raise_for_status()

        return { "file_upload": { "id": upload_id }, "name": basefilename }

    def multipart_upload(self, filename:str) -> dict:
        basefilename:str = os.path.basename(filename)
        size_left_of_sending_size: int = os.path.getsize(filename)
        number_of_parts:int = math.ceil(size_left_of_sending_size / self.chunk_size)

        data = {
            "mode": "multi_part",
            "filename": basefilename,
            "number_of_parts": number_of_parts,
        }
        upload_id:str = self.get_upload_id(data)

        headers:dict = self.api_handler.make_headers()
        endpoint_url:str = self.api_handler.make_endpoint_url(f"file_uploads/{upload_id}/send")
        for part_number in range(1, number_of_parts + 1):
            with open(filename, "rb") as f:
                sending_size:int = self.chunk_size if size_left_of_sending_size >= self.chunk_size else size_left_of_sending_size
                files = {
                    "file": (basefilename, f.read(sending_size)),
                    "part_number": (None, str(part_number))
                }

                response = requests.post(endpoint_url, headers=headers, files=files)
                response.raise_for_status()
                size_left_of_sending_size -= sending_size
            time.sleep(0.5)
        else:
            self.upload_completed(upload_id)

        return { "file_upload": { "id": upload_id }, "name": basefilename }

    def upload_completed(self, upload_id:str):
        headers:dict = self.api_handler.make_headers()
        endpoint_url = f"file_uploads/{upload_id}/complete"
        self.api_handler.post(endpoint_url, headers=headers)

    def retrieve_upload(self, upload_id:str) -> dict:
        headers:dict = self.api_handler.make_headers()
        endpoint_url = f"file_uploads/{upload_id}"
        return self.api_handler.get(endpoint_url, headers=headers)

    def import_external_file(self, external_url:str, filename:str) -> dict:   
        data = {
            "mode": "external_url",
            "filename": filename,
            "external_url": external_url
        }
        upload_id:str = self.get_upload_id(data)

        retry_limit:int = 5
        retry_interval:int = 2
        for retry_count in range(1, retry_limit + 1):
            upload_info:dict = self.retrieve_upload(upload_id)
            if upload_info["status"] == "uploaded":
                return { "file_upload": { "id": upload_id }, "name": filename }

            if retry_count < retry_limit:
                time.sleep(retry_interval ** retry_count)

        raise TimeoutError("Import imcomplete. operation is timed out.")
